/*
** Module   : SectionContent.java
** Abstract : Progress 4GL INI file content class representing one section of data
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130701 Created initial version.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * A class implementing the INI text file key=value secion lines. Can contain several lines with
 * key values and also comment lines end empty lines in any place of the section.
 *
 * @author    EVL
 */
public class SectionContent
extends Content
{
   /** Map to store key=value pars. */
   private Map<String, String> keyMap = new HashMap<String, String>();
   
   /** The section name associated with the content. */
   private String name = null;

   /** List to control section line ordering. */
   private List<Renderable> ordering = new LinkedList<Renderable>();
   
   /** The position of the last key=value pair in the ordering list. */
   private int nextKeyNumber = 0;

   /**
    * Constructs an instance of the data section in INI file.
    *
    * @param name 
    *        The name of the section corresponding to the key=value pairs of the INI file.
    */
   public SectionContent(String name)
   {
      super();
      this.name = name;
   }
   
   /**
    * Getting the name of this section.
    * 
    * @return The String with the section name.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Writes the content of the data section to the file via Writer object provided.
    * 
    * @param pw 
    *        The writer will be used to save content to the file.
    */
   public void write(PrintWriter pw)
   {
      if (ordering.size() > 0)
      {
         // Write the section name first
         pw.println("[" + name + "]");
         // Key names iterator
         Iterator iterLines = ordering.iterator();
         while (iterLines.hasNext())
         {
            // Get next available Renderable object
            Renderable lineToRender = (Renderable)iterLines.next();
            // And write them to the file
            String nextLine = lineToRender.render();
            // It is possible to have null pointer here if the key was previously removed
            if (nextLine != null)
            {
               pw.println(nextLine);
            }
         }
      }
   }
   
   /**
    * Checks if the content of the section is empty or not.
    * 
    * @return <code>true</code> in case of not emptry, <code>false</code> otherwise.
    */
   public boolean notEmpty()
   {
      // Not empty means we have some elements inside allocated map
      return ordering != null && ordering.size() > 0;
   }
   
   /**
    * Stores the key=value pair into the section internal map.
    * 
    * @param key 
    *        The String key name to store.
    * @param value 
    *        The String value to be associated with the given key name.
    */
   public void store(String key, String value)
   {
      // Find out if the key=value line already in the section
      boolean alreadyInMap = alreadyInSection(key); 
      // Put key=value pair into internal map 
      keyMap.put(key, value);
      // And refresh ordering list if the Renderer is still not there
      if (!alreadyInMap)
      {
         // We have to insert new key after last existed key, not at the end of the list
         // And increase then the next number
         ordering.add(nextKeyNumber++, new KeyRenderer(keyMap, key));
      }
   }

   /**
    * Removes the key=value pair from the section internal map. After this method the caller
    * must reload the INI file to sync up the internal maps. 
    * 
    * @param key 
    *        The String key name to remove.
    */
   public void remove(String key)
   {
      // Find out if the key=value line already in the section and remove them from map only
      if (alreadyInSection(key))
      {
         keyMap.remove(key);
      }
   }

   /**
    * Adds the key=value pair to the section internals assuming this is the new key.
    * 
    * @param key 
    *        The String key name to store.
    * @param value 
    *        The String value to be associated with the given key name.
    */
   public void add(String key, String value)
   {
      // Put key=value pair into internal map 
      keyMap.put(key, value);
      // And refresh ordering list if the Renderer is still not there
      ordering.add(new KeyRenderer(keyMap, key));
      // Remember the position of the next key to add in store() method
      nextKeyNumber = ordering.size();
   }

   /**
    * Stores the empty line or comment iside the section content.
    * 
    * @param textLine 
    *        The ignored text line to store.
    */
   public void store(String textLine)
   {
      // And refresh ordering map
      ordering.add(new TextRenderer(textLine));
   }
   
   /**
    * Gets the value associated with the given key inside the section.
    * 
    * @param key 
    *        The String key name of the value to be retrieved.
    * 
    * @return The String value associate with the given key or <code>null</code> if there is no
    *         key=value pair inside the section.
    */
   public String getValueForKey(String key)
   {
      if (keyMap != null)
      {
         return keyMap.get(key);
      }
      
      return null;
   }

   /**
    * Checks if the given key already added to the section internals..
    * 
    * @param key 
    *        The String key name to check.
    * 
    * @return <code>true</code> if the key was already added to internal map, <code>false</code>
    *         otherwise.
    */
   public boolean alreadyInSection(String key)
   {
      if (keyMap != null)
      {
         return keyMap.containsKey(key);
      }
      
      return false;
   }
   
   /**
    * Getting the comma-separated list of the keys for this section.
    * 
    * @return The String with comma-separated key names.
    */
   public String getKeys()
   {
      StringBuilder sb = new StringBuilder();
      
      Iterator iterLines = ordering.iterator();
      int count = 0;
      while (iterLines.hasNext())
      {
         // Get next available Renderable object
         Renderable lineToCheck = (Renderable)iterLines.next();
         // Get the string internal of the renderable, key or text members
         String keyToCheck = lineToCheck.toString();
         // And check if it is a key from key-value map 
         if (alreadyInSection(keyToCheck))
         {
            // This is the comma-separator if there is more than 1 name
            if (count > 0)
            {
               sb.append(EnvironmentOps.PROPATH_SEPARATOR);
            }
            sb.append(keyToCheck);
            count++;
         }
      }
      
      return sb.toString();
   }
   
   /**
    * The interface to present single line and key-value pair in one ordered list.
    */
   public interface Renderable
   {
      /**
       * Gets the string representation of the instance.
       * 
       * @return The String value associated with the given key=value line of the file or line
       *         to be ignored.
       */
      public String render();

      /**
       * Gets the string representation of the instance internal string part.
       * 
       * @return The String value associated with the given key or whole line.
       */
      public String toString();
   }

   /**
    * The class to present single line in one ordered list.
    */
   private static class TextRenderer
   implements Renderable
   {
      /** The content of the whole line. */
      private String text = null;

      /**
       * Constructs an instance of the single line in INI file.
       *
       * @param text 
       *        The text corresponding to the ignored line of the INI file.
       */
      private TextRenderer(String text)
      {
         this.text = text;
      }

      /**
       * Gets the string representation of the ignored line.
       * 
       * @return The String value associated with the given to be ignored.
       */
      public String render()
      {
          return text;
      }

      /**
       * Gets the string representation of the instance internal string part.
       * 
       * @return The String value associated with the given key or whole line.
       */
      public String toString()
      {
         return render();
      }
   }

   /**
    * The class to present key-value pair in one ordered list.
    */
   private static class KeyRenderer
   implements Renderable
   {
      /** The key name.*/
      private String key = null;

      /** The Map containing key=value pairs.*/
      private Map<String, String> lookup = null;

      /**
       * Constructs an instance of the key entry line in INI file.
       *
       * @param lookup 
       *        The Map containing the key-value pairs to look for.
       * @param key 
       *        The key part string of the given key=value pair.
       */
      private KeyRenderer(Map<String, String> lookup, String key)
      {
         this.lookup = lookup;
         this.key    = key;
      }

      /**
       * Gets the string representation of the key=value pair.
       * 
       * @return The String value associated with the given key=value line of the file.
       */
      public String render()
      {
         // Get the value for the known key
         String value = lookup.get(key);

         if (value != null)
            return key + "=" + value;
         else
            return null;
      }

      /**
       * Gets the string representation of the instance internal string part.
       * 
       * @return The String value associated with the given key or whole line.
       */
      public String toString()
      {
         return key;
      }
   }
}
